#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <assert.h>
#include <dirent.h>
#include <time.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <pthread.h>
#include <unistd.h>

#define streq(s,t) (strcmp (s, t) == 0)

#define LIBDIRPATH "./SOS"
//#define REPORT_INDIVIDUAL_TIMING 1

static struct timeval st, et;
static unsigned long aggrsec = 0;
static unsigned long aggrusec = 0;

static int 
begin_timer ()
{
  return (gettimeofday ( &st, NULL ));
}

static int 
time_stamp ( const char* description )
{
  int rc;

  if ( ! (timerisset (&st)) )
    return -1;

  rc = gettimeofday ( &et, NULL ); 

  if ( et.tv_usec >= st.tv_usec ) {
#if REPORT_INDIVIDUAL_TIMING 
    fprintf ( stdout, 
		"%s: %ld (%ld - %ld) seconds %ld (%ld - %ld) usec \n",
		description, 
		(et.tv_sec - st.tv_sec), et.tv_sec, st.tv_sec,
		(et.tv_usec - st.tv_usec), et.tv_usec, st.tv_usec);
#endif
    aggrsec += (et.tv_sec - st.tv_sec);
    aggrusec += (et.tv_usec - st.tv_usec);
  }
  else {
#if REPORT_INDIVIDUAL_TIMING  
    fprintf ( stdout, 
		"%s: %ld (%ld - %ld) seconds %ld (%ld - %ld) usec \n",
		description,
		( ( et.tv_sec - 1 ) - st.tv_sec), et.tv_sec, st.tv_sec,
		( (et.tv_usec+1000000) - st.tv_usec), et.tv_usec, st.tv_usec );
#endif
    aggrsec += ( ( et.tv_sec - 1 ) - st.tv_sec);
    aggrusec += ( (et.tv_usec+1000000) - st.tv_usec);
  } 

  return rc;
}

void *do_thread (void *arg)
{
  int *donep = (int *)arg;

  while (!*donep)
    ;

  return NULL;
}

#define MAX_THREADS 512

int main(int argc, char* argv[])
{
  struct dirent *e;
  struct stat st;
  DIR *dirp;
  void *plug;
  char sopath[1024] = {0};
  char msg[1024] = {0};
  int i;
  pthread_t threads[MAX_THREADS]; 
  void *exit_vals[MAX_THREADS];
  int done = 0;
  long long nthreads = 3;

  for (i = 1; i < argc; i++)
    {
      if (streq (argv[i], "-nthreads"))
	{
	  i++;
	  nthreads = strtoll (argv[i], (char **)0, 0);
	}
    }

  for (i = 0; i < nthreads; i++)
    {
      pthread_create(&threads[i], NULL, do_thread, &done);
    }

  dirp = opendir( LIBDIRPATH );

  if ( dirp == NULL ) {
    fprintf ( stderr, "[DLOPEN TEST] Error opening %s\n", LIBDIRPATH);
    return EXIT_FAILURE;
  }

  for ( ; ; ) {
     e = readdir( dirp );

     if ( e == NULL )
       break;
     sprintf (sopath, "%s/%s", LIBDIRPATH, e->d_name);

     /* Check only regular files. */
     if ( (strncmp(e->d_name, ".", 1) == 0)
	  ||   (stat( sopath, &st ) < 0)
	  ||   (! S_ISREG(st.st_mode)) )
       continue;

     sprintf (msg, "[DLOPEN TEST] Time to dlopen %s\n", sopath);

     begin_timer ();
     plug = dlopen (sopath, RTLD_LAZY);    
     time_stamp (msg);
     
     if (plug == NULL) {
       dlerror ();
       continue;
     }
     
     sprintf (msg, "[DLOPEN TEST] Time to dlclose %s\n", sopath);

     begin_timer ();
     dlclose (plug);         
     time_stamp (msg);
  }

  closedir (dirp);

  if (aggrusec > 1000000UL) {
    unsigned long carry = aggrusec/1000000;
    aggrsec += carry;
    aggrusec -= carry*1000000;  
  }

  double sec = (double) aggrsec + ((double)aggrusec / 1000000.0);

  fprintf (stdout, "Aggregate Time: %f sec\n", sec);

  done = 1;

  for (i = 0; i < nthreads; i++)
    {
      pthread_join(threads[i], &exit_vals[i]);
    }

  return EXIT_SUCCESS;
}
